################################################################################
#
#       This is part of CATBox (Combinatorial Algorithm Toolbox)
#       version 1.2.7 from 1/25/2021. You can find more information at
#       http://schliep.org/CATBox/
#
#	file:   PrimalDualKruskal.pro
#	author: Alexander Schliep (alexander@schliep.org)
#
#       Copyright (C) 1998-2020, Winfried Hochstaettler and Alexander Schliep.
#	(C) 2010 Springer Verlag.
#       
#	This is part of the Springer textbook "CATBox - an interactive course in 
#       discrete optimization".
#
#       All rights reserved. Do not distribute without written permission. 
#
#
#
#
################################################################################


#### Options #################################################

breakpoints = [1]
interactive = []
graphDisplays = 2
about = """<HTML>
<HEAD>
<TITLE>Minimal Spanning Tree - Kruskal (Primal / Dual)</TITLE>
</HEAD>
<BODY>

<H1>Minimal Spanning Tree - Kruskal (Primal / Dual)</H1>

This algorithm finds a minimal spanning tree. Edges in the input are
disregarded. Vertices in the input are interpreted as points in the
Euclidean plane and the resulting MST can contain edges not present
in the input.

<H3>Visualization (Main Graph)</H3>

<H5>Moat Colors</H5> <dl> <dt><colordef color="#FF0000"><colordef
color="#FF8800"><colordef color="#FFFF00">...</dt> <dd> Moats are
graphical representations of dual variables for individual vertices or
groups of vertices. The growing of moats visualizes increases in the
value of the corresponding dual variables.</dd>

</dl>


<H5>Edge Colors</H5>
<dl>

<dt><colordef color="#EEEEEE"></dt> <dd>Edges in this color are
part of the MST.</dd>

</dl>


<H3>Visualization (MST)</H3>

<H5>Vertex Colors</H5>

<dl>

<dt><colordef color="red"><colordef color="green"><colordef color="blue">...</dt> <dd>Vertex (and edge) colors designate the connected
components in the subgraph induced by the trees during the
computation. Therefore, each vertex has its own colors initially.</dd>

</dl>



<H5>Edge Colors</H5>

<dl>

<dt><colordef color="red"><colordef color="green"><colordef color="blue">...</dt> <dd>Edges
in the (partial) MST; color indicates the connected component.</dd>



</dl>




</BODY></HTML>
"""
#### Misc ####################################################

from math import sqrt

for e in G.Edges():
    G.DeleteEdge(e[0],e[1])
G.directed = 0

import copy
GS = copy.deepcopy(G)
self.OpenSecondaryGraph(GS, stripPath(self.graphFileName) + " (spanning)")
GSA = self.GUI.secondaryGraphDisplay
A.ShowGraph(G, stripPath(self.graphFileName))


class Moat:
    def __init__(self, graph_display, pos, embed_radius, color, vertex_num, moat_id):
        d = (embed_radius * graph_display.zoomFactor ) / 100.0 #Canvas radius!
        self.d = d
        self.moat_id = moat_id
        self.vertex_num = vertex_num
        self.pos = pos
        self.graph_display = graph_display
        m = self.graph_display.canvas.create_oval(pos[0]-d, pos[1]-d, pos[0]+d, pos[1]+d, 
            fill=color, tag="moats", width=0)
        self.graph_display.CreateMoat(moat_id, d, color) # Get the animation in history
        self.m = m
        self.graph_display.canvas.lower(m, "vertices")
        try:
            self.graph_display.canvas.lower(m, "moats")
        except:
            pass

    def grow(self, new_embed_radius):
        # XXX Don't think we need this in GraphDisplay right now
        d = (new_embed_radius * self.graph_display.zoomFactor ) / 100.0 #Canvas radius!
        self.graph_display.canvas.coords(self.m,
                             self.pos[0]-d, self.pos[1]-d,
                             self.pos[0]+d, self.pos[1]+d)

class MoatedPartition:
    """ A Moated partition is a collection of connected components
        where each vertex in a component has a moat around it.
        The moated partition keeps track of distances between vertices """
    def __init__(self, graph, secondary_graph, graph_display, secondary_graph_display):
        self.graph = graph
        self.secondary_graph = secondary_graph
        self.graph_display = graph_display
        self.secondary_graph_display = secondary_graph_display
        
        #self.moatRadius = {} # Radius of the moat per vertex
        #for v in self.graph.vertices:
        #    self.moatRadius[v] = 0.0

        self.mr = 0

        self.D = {} # Dictionary of distance d(u,v). Assume u < v because of symmetry
        for u in self.graph.vertices:
            for v in self.graph.vertices:
                if u < v:
                    tmp = (self.graph.embedding[u].x - self.graph.embedding[v].x)**2
                    tmp += (self.graph.embedding[u].y - self.graph.embedding[v].y)**2
                    self.D[(u,v)] = max(sqrt(tmp), 0.0) # Enforce non-negative distances

        self.cm = ComponentMaker(self.graph, self.secondary_graph_display)
        self.components = []
        self.component = {} 
        for u in self.graph.vertices:
            C = self.cm.NewComponent()
            C.AddVertex(u)
            self.components.append(C)
            self.component[u] = C

        self.colorList = ["#FF0000","#FF8800","#FFFF00","#00FF00","#00FFFF","#0088FF",
                          "#0000FF","#8800FF","#FF00FF"]
        self.colorIndex = -1
        

    def distance(self, u, v):
        if u < v:
            return self.D[(u,v)]
        else:
            return self.D[(v,u)]


    def findMinimalDistance(self):
        """ find the (u,v) having the minimal positive entry in self.D """
        def swapTupel(t):
            return (t[1],t[0])

        # items() gives (key, value)-List, swap yields (value, key), and max works on that
        (d, e) = min(filter(lambda t: t[0] > 0.0, map(swapTupel,self.D.items())))
        return (e,d)


    def NextColor(self):
        if self.colorIndex == len(self.colorList) - 1:
            self.colorIndex = - 1
        self.colorIndex = self.colorIndex + 1
        return self.colorList[self.colorIndex]
       
        
    def GrowMoats(self,dist):
        if not getattr(self, 'grow_round', None):
            self.grow_round = 0
        moatColor = self.NextColor()

        moats = []

        # No point animating growing moats behind vertices
        if self.mr == 0:
            old_mr = A.zVertexRadius
        else:
            old_mr = self.mr
        self.mr += dist/2.0
       
        for C in self.components:
            for v in C.vertices:
                pt = self.graph_display.G.GetEmbedding(v)  
                x,y = self.graph_display.EmbeddingToCanvas(pt.x,pt.y)
                m = Moat(self.graph_display, (x,y), old_mr, moatColor, v, 'moat_' + str(v) + '-' + str(self.grow_round)) 
                moats.append(m)

        stepwidth = 0.5 # old value was 0.05
        steps = int(round((self.mr - old_mr)/stepwidth))
        delta = (self.mr - old_mr) / steps
        
        for m in moats:
            self.graph_display.GrowMoat(m.moat_id, old_mr + steps*delta)
        for i in range(1, steps + 1): 
            for m in moats:
                m.grow(old_mr + i*delta - 0.001) # constant is fudge factor to let moats not quite touch
            self.graph_display.update()
        self.graph_display.Wait()

        self.tightEdges = self.UpdateExtraComponentDistances(dist)
        self.grow_round += 1


    def UpdateComponents(self):
        for e in self.tightEdges: # self.tightEdges is set by GrowMoats. Call first!
            (u,v) = e
            if self.component[u] is not self.component[v]:
                self.graph_display.AddEdge(u,v)
                srcComp = self.component[v]
                self.component[u].AddSubGraph(srcComp)
                self.component[u].AddEdge(u,v)
                for v in srcComp.vertices:
                    self.component[v] = self.component[u]
                self.components.remove(srcComp)

        self.UpdateIntraComponentDistances()


    def UpdateExtraComponentDistances(self, dist):
        tightEdges = []
        growingFactor = dist / 2
        for C in self.components:
            for D in self.components:
                if C is not D:
                    for u in C.vertices:
                        for v in D.vertices:
                            if u < v:
                                e = (u,v)
                            else:
                                e = (v,u)
                            self.D[e] = self.D[e] - growingFactor
                            if self.D[e] == 0.0:
                                tightEdges.append(e)
                                self.secondary_graph_display.AddEdge(e[0],e[1])
                                    
        return tightEdges

    def UpdateIntraComponentDistances(self):
        for C in self.components:
            for u in C.vertices:
                for v in C.vertices:
                    if u is not v:
                        self.D[(u,v)] = 0.0
          
MP = MoatedPartition(G, GS, A, GSA)

nrOfComponents = lambda mp=MP: len(mp.components)
UpdateDuals = lambda mp=MP: mp.GrowMoats(mp.findMinimalDistance()[1])
ComputePartition = lambda mp=MP: mp.UpdateComponents()
