################################################################################
#
#       This is part of CATBox (Combinatorial Algorithm Toolbox) 
#       version 1.2.7 from 1/25/2021. You can find more information at 
#       http://schliep.org/CATBox/
#
#       file:   FordFulkerson.pro
#       author: Torsten Pattberg (torsten.pattberg@klf.de)
#               Alexander Schliep (alexander@schliep.org)
#
#       Copyright (C) 1998-2020, Winfried Hochstaettler and Alexander Schliep.
#	(C) 2010 Springer Verlag.
#       
#	This is part of the Springer textbook "CATBox - an interactive course in 
#       discrete optimization".
#
#       All rights reserved. Do not distribute without written permission. 
#
#
#
#
################################################################################

#### Options #################################################

breakpoints = [14,17]
interactive = [9,10]
graphDisplays = 2
self.NeededProperties({'Directed':1, 'EdgeWeights':1, 'NonNegativeEdgeWeights':1})
about = """<HTML>
<HEAD>
<TITLE>Maximal Flow - FordFulkerson</TITLE>
</HEAD>
<BODY>
</BODY></HTML>
"""

#### Wrapper #################################################


class BlinkingEdges(BlinkingContainerWrapper):

    def __init__(self, theAnimator, l):
        BlinkingContainerWrapper.__init__(self, theAnimator, l, "black")

    def __getitem__(self, i):
        if i < len(self.list):
            item = self.list[i]
            self.Animator.BlinkEdge(item[0],item[1],"gray")
            return item
        else:
            raise IndexError

#### Graph management ########################################

import copy
R = copy.deepcopy(G)
self.OpenSecondaryGraph(R, stripPath(self.graphFileName) + " (residual)",
                        WeightedGraphInformer(R,"residual capacity"))
RA = self.GUI.secondaryGraphDisplay
G.CalculateWidthFromWeight(0.7)
A.ShowGraph(G, stripPath(self.graphFileName))

#### Variables ###############################################

s       = 1         # first vertex is source
t       = G.Order() # last vertex is sink
true    = 1
maximal = 0
flow   = FlowWrapper(G,A,R,RA,G.edgeWeights[0],R.edgeWeights[0])

cap         = lambda e: flow.cap[e]
res         = lambda e: flow.res[e]
ForwardEdge = lambda u,v,g=G: g.QEdge(u,v)

def pickCallback(A,RA,v,type):
    if type == "s":
        A.SetVertexAnnotation(v,type,"green")
        flow.excess[v] = gInfinity
        RA.SetVertexColor(v,"green")
    else:
        A.SetVertexAnnotation(v,type,"red")
        flow.excess[v] = -gInfinity
        RA.SetVertexColor(v,"red")

pickCallbackSource = lambda v, type="s", a=A,ra=RA: pickCallback(a,ra,v,type)
PickSource         = lambda f=pickCallbackSource: self.PickVertex(1,None,f)
pickCallbackSink   = lambda v, type="t", a=A,ra=RA: pickCallback(a,ra,v,type)
PickSink           = lambda f=pickCallbackSink: self.PickVertex(G.Order(),None,f)

#### Internal functions ######################################

def _MinResCap(Path, R, RA):
    min = gInfinity
    u   = 0
    v   = 0
    for i in Path:
        if res(i) < min:
            min = res(i)
            u = i[0]
            v = i[1]

    RA.SetEdgeColor(u,v,"#FFAAAA") # Light the cheapest edge

    if R.edgeWeights[0].QInteger():
        RA.graphInformer.SetDefaultInfo("delta = %d"%min)
        RA.UpdateInfo("delta = %d"%min)
    else:
        RA.graphInformer.SetDefaultInfo("delta = %.2f"%min)
        RA.UpdateInfo("delta = %.2f"%min)

    return min


def _ShowCut(A,R,s):
    Q = Queue() # Component S of s is green, component T of t is red
    C = [s]     # edges between S and T are yellow

    pred = {}
    for v in R.vertices:
        pred[v] = None
    Q.Append(s)

    while Q.IsNotEmpty():
        v = Q.Top()
        for w in R.Neighborhood(v):
            if pred[w] == None and w != s:
                pred[w] = v
                Q.Append(w)
                C.append(w)

    A.SetAllVerticesColor("red")

    for v in C:
        A.SetVertexColor(v,"green")
        for w in G.Neighborhood(v):
            if not w in C:
                A.SetEdgeColor(v,w,"yellow")

    return None

#### Lambdas #################################################

ShortestPath    = lambda s,t,r=R,ra=RA:shortestPath(r,ra,s,t)
Neighborhood    = lambda v,r=R: r.Neighborhood(v)
ShowCut         = lambda s,a=A,r=R: _ShowCut(a,r,s)
MinResCap       = lambda p,ra=RA,r=R: _MinResCap(p,r,ra)
Edges           = lambda r, ra=RA: BlinkingEdges(ra,r)

#### Internal initialisation #################################

A.SetAllVerticesColor("grey")
RA.SetAllVerticesColor("grey")

A.RegisterGraphInformer(FlowGraphInformer(G,flow))
RA.RegisterGraphInformer(ResidualGraphInformer(R,flow))
