################################################################################
#
#       This is part of CATBox (Combinatorial Algorithm Toolbox)
#       version 1.2.7 from 1/25/2021. You can find more information at
#       http://schliep.org/CATBox/
#
#	file:   NegativeCircuit.pro
#	author: Torsten Pattberg (torsten.pattberg@klf.de)
#               Alexander Schliep (alexander@schliep.org)
#
#       Copyright (C) 1998-2020, Winfried Hochstaettler and Alexander Schliep.
#	(C) 2010 Springer Verlag.
#       
#	This is part of the Springer textbook "CATBox - an interactive course in 
#       discrete optimization".
#
#       All rights reserved. Do not distribute without written permission. 
#
#
#
#
################################################################################

#### Options #################################################

breakpoints = [5,9]
interactive = []
graphDisplays = 2
self.NeededProperties({'EdgeWeights':2, 'Directed':1})
about = """<HTML>
<HEAD>
<TITLE>Minimum Cost Flow - Negative Cycle</TITLE>
</HEAD>
<BODY>

This algorithm finds a minimum cost flow by canceling negative
cycles. Initially, adding an artificial source-sink-pair and using a
maximum flow algorithm, a feasible flow is computed. In the following
we search for cycles of negative weight in the residual network, and
decrease the cost by sending flow along these cycles, see Chapter 7 of
the book.


<H3>Visualization 
 </H3>

<H4>
Primary Window
</H4>

<H5>Vertex Colors</H5> <dl>
<dt><colordef color=green></dt> <dd>
Vertices with a positive supply.</dd> 

<dt><colordef color="red"></dt> <dd>
Vertices with a positive demand.</dd> 

<dt><colordef color="grey"></dt> <dd>
Balanced vertices.</dd> 
 </dl>


<H5>Edge Colors</H5> <dl><dt><colordef color="#EEEEEE"></dt>
Initial color for edges. </dd>
<dt><colordef color="blue"></dt> <dd>Color of saturated edges.</dd> 
<dt><colordef color="#9999FF"></dt> <dd>Color of
unsaturated edges with positive flow.</dd>
</dl>

<H5>Further elements</H5> <dl><dd> Labels at the vertices give the
suppy respectively the demand of the vertex. Labels at edges give the
flow and the capacity of that edge.</dd> </dl>
 </dl>

<H4>
<a name="Secondary"> Secondary Window</a>
</H4>

This window depicts the residual network. The residual capacities and
the cost of the edges are given in the graph informer.

<p>
<p>

<H5>Vertex Colors</H5> <dl>
<dt><colordef color=green></dt> <dd>
Vertices with a positive supply.</dd> 

<dt><colordef color="red"></dt> <dd>
Vertices with a positive demand.</dd> 

<dt><colordef color="grey"></dt> <dd>
Balanced vertices.</dd> 

<dt><colordef color="red"></dt> <dd>
Vertices in a negative cycle.</dd> 

<dt><colordef color="green"></dt> <dd>
Vertex processed in an augmentation along a negative cycle.
</dd></dl>
<p>

<H5>Edge Colors</H5> 
<dl><dt><colordef color="#EEEEEE"></dt>
Initial color for edges. </dd>
<dl><dt><colordef color="red"></dt>
Edges in a negative cycle. </dd>

<dt><colordef color="yellow"></dt> <dd> Edge of minimum capacity in
the present negative cycle.  
<dt><colordef color="green"></dt> <dd>
Edges in the shortest path tree proving the non-existence of a negativ cycle.
</dd></dl>
<H5>Further elements</H5> <dl><dd> Labels at the vertices give the dually optimal potential.</dd> </dl>
 





</BODY></HTML> """
self.NeededProperties({'EdgeWeights':2, 'VertexWeights':1})

#### Misc ####################################################

from math import *
import pprint
ready = 0

#### Graph management ########################################

import copy

R  = copy.deepcopy(G)
self.OpenSecondaryGraph(R, stripPath(self.graphFileName) + " (residual network)")
RA = self.GUI.secondaryGraphDisplay
G.CalculateWidthFromWeight(0.7)
A.ShowGraph(G, stripPath(self.graphFileName))

#### Wrapper #################################################

class MyFlowWrapper(FlowWrapper):
    """
        Set flow to zero without changing the excess. Necessary when a new
        edge is added after initalizing the flow
    """

    def ForceZero(self,e):
        self.flow[e] = 0
        return



#### GraphInformer ###########################################

class MyResidualGraphInformer(ResidualGraphInformer):
    """ Also show the costs and some info about the cycle
    """
    def EdgeInfo(self,v,w):
        if not C or (v,w) not in C.Edges():
            return "Edge (%d,%d) - residual capacity: %d - cost: %d" % (v, w, self.res[(v,w)],cost[(v,w)])
        else:
            cyclecost = 0
            for e in C.Edges():
                    cyclecost = cyclecost + cost[e]
            return "Edge (%d,%d) - residual capacity: %d - cost: %d - Cycle cost: %d" % (v, w, self.res[(v,w)],cost[(v,w)],cyclecost)

#### Variables ###############################################

cost = G.edgeWeights[1]
flow = MyFlowWrapper(G,A,R,RA,G.edgeWeights[0],R.edgeWeights[0],G.vertexWeights[0])

pred = {}
dist = {}
mark = {}

NOW  = Queue()
NEXT = Queue()

D    = Stack()

#### Lambdas #################################################

C = AnimatedSubGraph(R,RA)

NewCycle              = lambda r=R, ra=RA: AnimatedSubGraph(r,ra,"red")

EstablishFeasibleFlow = lambda ra=RA,a=A: EstablishFeasibleFlow_(ra,a)
Vertices              = lambda r=R: r.vertices
Order                 = lambda g=G: g.Order()
Edges                 = lambda l, ra=RA: BlinkingContainerWrapper(ra,l.Edges(),"black")

cap                   = lambda e: flow.cap[e]
excess                = lambda v: flow.excess[v]
res                   = lambda e: flow.res[e]
ForwardEdge           = lambda u,v,g=G: g.QEdge(u,v)

#### Internal functions ######################################

def EstablishFeasibleFlow_(RA,A):
    bb = A.canvas.bbox('all')

    delta = 0
    
    A.AddVertex(bb[0] + bb[2] + delta, bb[1] + bb[3] + delta)
    A.AddVertex(bb[0] - delta, bb[1] - delta)
    sink   = RA.AddVertex(bb[0] + bb[2] + delta, bb[1] + bb[3] + delta)
    source = RA.AddVertex(bb[0] - delta, bb[1] - delta)

    flow.excess[source] = gInfinity
    flow.excess[sink]   = -gInfinity

    A.SetAllVerticesColor("grey")
    A.SetVertexAnnotation(source,"source","green")
    A.SetVertexAnnotation(sink,"sink","red")
    A.SetVertexColor(source,"green")
    A.SetVertexColor(sink,"red")
    RA.SetVertexColor(source,"green")
    RA.SetVertexColor(sink,"red")

    maxWeight = max(R.edgeWeights[0].label.values()) #XXX Using G internals
    #print R.edgeWeights[0].label
    #print "MaxWeight", maxWeight
    
    for v in Vertices():

        if (v != sink) and (v != source):
            if excess(v) > 0:
                # XXX Modifying G internals 
                G.edgeWidth[(source,v)]  = 0.7 * (1 + 35 * excess(v) / maxWeight)
                
                A.AddEdge(source, v)
                RA.AddEdge(source, v)

                flow.cap[(source,v)] = excess(v)
                flow.ForceZero((source,v))
                cost[(source,v)] = 0
                cost[(v,source)] = 0
                flow.res[(source,v)] = excess(v)
                flow.res[(v,source)] = 0
                A.SetVertexColor(v,"green")
                A.SetVertexAnnotation(v,excess(v),"green")
            elif excess(v) < 0:
                # XXX Modifying G internals 
                G.edgeWidth[(v,sink)] = 0.7 * (1 - 35 * excess(v) / maxWeight)

                A.AddEdge(v, sink)
                RA.AddEdge(v, sink)
                flow.cap[(v,sink)] = -excess(v)
                flow.ForceZero((v,sink))
                cost[(v,sink)] = 0
                cost[(sink,v)] = 0
                flow.res[(v,sink)] = -excess(v)
                flow.res[(sink,v)] = 0
                A.SetVertexColor(v,"red")
                A.SetVertexAnnotation(v,excess(v),"red")
                D.Push(v)

    Cmax = 0
    for e in R.Edges():
        if (cap(e) > Cmax):
            Cmax = cap(e)

    k = pow(2, floor(log(Cmax)/log(2)))

    while k >= 1:

        while not ready:

            P = FindPath(R,RA,source,sink,k)

            if P == None:
                break
            else:
                delta = MinimalRestCapacity2(P)

                for (u,v) in P.Edges():
                    if ForwardEdge(u,v):
                        flow[(u,v)] = flow[(u,v)] + delta
                    else:
                        flow[(v,u)] = flow[(v,u)] - delta
        k = k / 2

    for v in R.Neighborhood(source):
        if res((source,v)):
            #print "ERROR: No feasible flow found!"
            return False

    A.DeleteVertex(source)
    A.DeleteVertex(sink)
    RA.DeleteVertex(source)
    RA.DeleteVertex(sink)

    return True


def FindPath(R,A,s,t,k):

    P    = SubGraph(R)
    Q    = Queue()

    pred = {}

    for v in R.vertices:
        pred[v] = None

    Q.Append(s)

    while Q.IsNotEmpty():

        v = Q.Top()

        for w in R.Neighborhood(v):

            if pred[w] == None and w != s and res((v,w)) >= k:

                pred[w] = v
                Q.Append(w)

                if w == t:
                    while pred[w] != None:
                        P.AddEdge(pred[w],w)
                        w = pred[w]
                    return P

    return None



def ShowTree(R,RA):

    for v in R.vertices:
        RA.SetVertexAnnotation(v,dist[v],"black")
        try:
            RA.SetEdgeColor(pred[v],v,"green")
        except:
            pass


def MinimalRestCapacity2(Graph):
    min = gInfinity
    for i in Graph.Edges():
        if res(i) < min:
            min = res(i)

    return min


def MinimalRestCapacity(Graph):
    min = gInfinity
    u   = 0
    v   = 0
    for i in Graph.Edges():
        if res(i) < min:
            min = res(i)
            u = i[0]
            v = i[1]

    RA.SetEdgeColor(u,v,"yellow")

    return min

#### Internal initialisation #################################
for e in G.Edges():
    cost[(e[1],e[0])] = -cost[e]

RA.SetAllVerticesColor("grey")

A.RegisterGraphInformer(FlowGraphInformer(A,flow))
RA.RegisterGraphInformer(MyResidualGraphInformer(R,flow))

#### functions ############################

def FindNegativeCycle():
    RA.SetAllVerticesColor("#EEEEEE")
    RA.SetAllEdgesColor("#EEEEEE")

    while D.IsNotEmpty():

        v = D.Pop()
        C = SearchCycleFrom(v)
        if not C == None:
            D.Push(v)
            return C

    return None

def SearchCycleFrom(v):
    NC = NewCycle()

    NOW.Clear()
    NEXT.Clear()

    for u in Vertices():
        dist[u] = gInfinity
        pred[u] = None
        mark[u] = False

    NOW.Append(v)
    dist[v] = 0
    counter = 0

    while not ready:
        if NOW.IsEmpty():
            if NEXT.IsEmpty():
                return None
            else:
                counter = counter + 1

                if counter > Order():
                    break

                while NEXT.IsNotEmpty():
                    u = NEXT.Top()
                    NOW.Append(u)

        else:
            u = NOW.Top()

            for w in R.Neighborhood(u):

                if dist[w] > dist[u] + cost[(u,w)]:
                    dist[w] = dist[u] + cost[(u,w)]
                    pred[w] = u

                    if not (NOW.Contains(w) or NEXT.Contains(w)):
                        NEXT.Append(w)

    while not mark[w]:
        mark[w] = True
        w = pred[w]

    first_vertex = w

    while not (pred[w] == first_vertex):
        NC.AddEdge(pred[w],w)
        w = pred[w]

    NC.AddEdge(pred[w],w)

    return NC

# Min Mean Cycle canceling
def MinMeanCycle():
    "Find minimum mean weight circuit in a weighted digraph."
    RA.SetAllVerticesColor("#EEEEEE")
    RA.SetAllEdgesColor("#EEEEEE")

    while D.IsNotEmpty():
        v = D.Pop()
        C = SearchMinMeanCycleFrom(v)
        if C != None:
            D.Push(v)
            return C

    return None

def SearchMinMeanCycleFrom(root):
    dist2 = {}
    pred2 = {}
    n = Order()

    for u in Vertices():
        for k in range(0, n+1):
            dist2[(k, u)] = gInfinity
            pred2[(k, u)] = None

        mark[u] = False

    dist2[(0, root)] = 0

    # Compute the distance and predecessor table
    for k in range(1, n+1):
        for u in Vertices():
            for w in R.Neighborhood(u):
                if dist2[(k, w)] > dist2[(k-1, u)] + cost[(u, w)]:
                    dist2[(k, w)] = dist2[(k-1, u)] + cost[(u, w)]
                    pred2[(k, w)] = u

    # Find the minimum cycle mean.
    minMean = gInfinity
    minV = None
    for u in Vertices():
        maxMean = -gInfinity
        for k in range(0, n):
            maxMean = max(maxMean, float(dist2[(n, u)] - dist2[(k, u)]) / (n - k))

        if minMean > maxMean:
            minMean = maxMean
            minV = u

    if minMean >= 0:
        # There is no negative-weight cycle.
        # Compute a shortest-path tree (in dist and pred) using the
        # Bellman-Ford algorithm as a certificate.
        for v in Vertices():
            dist[v] = gInfinity
            pred[v] = None

        dist[root] = 0
        count = 0

        NEXT.Append(root)
        while count < Order() and NEXT.IsNotEmpty():
            if NOW.IsEmpty():
                while NEXT.IsNotEmpty():
                    v = NEXT.Top()
                    NOW.Append(v)
                count += 1
            while NOW.IsNotEmpty():
                v = NOW.Top()
                for w in R.Neighborhood(v):
                    if dist[w] > dist[v] + cost[(v,w)]:
                        dist[w] = dist[v] + cost[(v,w)]
                        pred[w] = v
                        if not NEXT.Contains(w):
                            NEXT.Append(w)

        return None

    # There is a negative-weight cycle. A min-mean circuit can now be
    # found by backtracking from minV.
    w = minV
    k = n
    while not mark[w]:
        mark[w] = True
        w = pred2[(k, w)]
        k -= 1

    # Found a cycle starting at w. Backtrack again from minV to get
    # the correct k to use for indexing into pred2.
    firstVertex = w
    w = minV
    k = n
    while w != firstVertex:
        w = pred2[(k, w)]
        k -= 1

    NC = NewCycle()

    while pred2[(k, w)] != firstVertex:
        NC.AddEdge(pred2[(k, w)], w)
        w = pred2[(k, w)]
        k -= 1

    NC.AddEdge(pred2[(k, w)], w)

    return NC
