################################################################################
#
#       This is part of CATBox (Combinatorial Algorithm Toolbox)
#       version 1.2.7 from 1/25/2021. You can find more information at
#       http://schliep.org/CATBox/
#
#   file:   Bipartite.pro
#   author: Torsten Pattberg (torsten.pattberg@klf.de)
#               Alexander Schliep (alexander@schliep.org)
#
#       Copyright (C) 1998-2020, Winfried Hochstaettler and Alexander Schliep.
#	(C) 2010 Springer Verlag.
#       
#	This is part of the Springer textbook "CATBox - an interactive course in 
#       discrete optimization".
#
#       All rights reserved. Do not distribute without written permission. 
#
#
#
################################################################################

#### Options #################################################

breakpoints = [44]
interactive = []
graphDisplays = 1
self.NeededProperties({'Bipartite':1, 'Undirected':1})
about = """<HTML>
<HEAD>
<TITLE>Matching - Bipartite</TITLE>
</HEAD>
<BODY>

This algorithm finds a maximum matching in a bipartite graph with
partition V (red) and W (green). To proove the result it will show a
minimum edge cover that has (in bipartite graphs) the same cardinality
as the maximum matching (see Chapter 8 of the book).

<H3>Visualization 
 </H3>

<H5>Vertex Colors</H5> <dl>
<dt><colordef color="#ff9999"></dt> <dd>
Initial color of the vertices. Color of one color class when the bipartition has been computed.</dd> 
<dt><colordef color="#99ff99"></dt> <dd>
Color of the other color class when the bipartition has been computed.</dd> 
<dt><colordef color="#00cc00"></dt> <dd> Color of vertices that are in
the queue Q or have been removed from the queue in the present
procedure to search for an augmenting path.</dd>
<dt><colordef color=red></dt> <dd>
Vertices of the red color class that have been marked in the present
procedure to search for an augmenting path.</dd>
<dt><colordef color="black"></dt> <dd>The vertex being <em>processed</em> is
displayed with a wide outline in this color.</dd>
 </dl>
When a maximum matching has been found we use the following colors 
show the vertex cover.
<dl>
<dt><colordef color=red></dt> <dd>
<em> Covering </em> vertices from the red color class.</dd> 
<dt><colordef color="#99ff99"></dt> <dd>
Covering vertices from the green color class.</dd> 
<dt><colordef color="black"></dt> <dd>All covering vertices
are displayed with a wide outline in this color.</dd>
</dl>

<p>

<H5>Edge Colors</H5> <dl><dt><colordef color="#EEEEEE"></dt>
Initial color for all edges in the very beginning. </dd>
<dt><colordef color="black"></dt>
Initial color for all edges when a search forest is grown. </dd>
<dt><colordef color="blue"></dt> <dd>Edges in a matching.</dd>
<dt><colordef color="yellow"></dt> <dd>Edge that is currently processed.</dd>
<dt><colordef color="#ccccff"></dt> <dd>Non-matching edges in the search forest.</dd>
<dt><colordef color="grey"></dt> <dd>Edges that have been processed but are not in the search forest.</dd>
</dl>


<H5>Further elements</H5> <dl> <dt><colordef color="yellow"></dt> <dd>
When an augmenting path is found we indicate this by a yellow shade around the
edges of the path.</dd>
</dl>


</BODY></HTML>
"""

#### Wrapper #################################################

class MatchingGraphInformer(GraphInformer):

    def EdgeInfo(self,tail,head):
        if (tail,head) in M.Edges():
            return "Edge (%d,%d) is a matching edge"%(tail, head)
        else:
            return "Edge (%d,%d) is not a matching edge"%(tail, head)

    def VertexInfo(self,v):
        if v in V.vertices:
            set = "V"
        else:
            set = "W"

        if pred[v] == v:
            info = " is exposed"
        elif pred[v] != None:
            info = " has predecessor %s"%pred[v]
        else:
            info = " has no predecessor"

        return "Vertex %d in %s %s"%(v, set, info)

#### Graph management ########################################

M = AnimatedSubGraph(G, A, "blue")
V = AnimatedSubGraph(G, A, "#99ff99")
W = AnimatedSubGraph(G, A, "#ff9999")
Q = AnimatedVertexQueue(A, "#00cc00", "#00cc00")

class MyAnimatedPredecessor(AnimatedPredecessor):

    def __setitem__(self, v, val):
        try:
            oldVal = VertexLabeling.__getitem__(self, v)
            if oldVal != None:
                if self.leaveColors == None or not (self.Animator.GetEdgeColor(oldVal,v) in self.leaveColors):
                    self.Animator.SetEdgeColor(oldVal,v,"#EEEEEE")
        except:
            pass

        if val != None:
            try:
                if v in V.vertices:
                    self.Animator.SetVertexColor(v,"#00cc00")
                    self.Animator.SetEdgeColor(val,v,"blue")
                else:
                    self.Animator.SetVertexColor(v,"red")
                    self.Animator.SetEdgeColor(val,v,"#ccccff")
            except:
                pass
            else:
                if v in V.vertices:
                    self.Animator.SetVertexColor(v,"#99ff99")
                else:
                    self.Animator.SetVertexColor(v,"#ff9999")
        
        VertexLabeling.__setitem__(self, v, val)


#### Variables ###############################################

dist = {}
pred = MyAnimatedPredecessor(A,["blue"])

#### Internal functions ######################################


def DeleteLabels():
    Q.Clear()
    for v in G.vertices:
        pred[v] = None
    A.SetAllEdgesColor("black", None, ["blue"])


highlightID = None

def ShowPath(P):
    global highlightID
    if highlightID:
        A.HidePath(highlightID)
        highlightID = None
    if P:
        # Fix orientations of edges
        if len(P) > 1:
            for i, e in enumerate(P[:-1]):
                if e[0] in P[i+1]: # (u,v),(v,w)
                    P[i] = e[1], e[0]
            if P[-1][1] in P[-2]:
                P[-1] = (P[-1][1], P[-1][0])
        V = [P[0][0]] + [v for u,v in P]
        #print P, V
        highlightID = A.HighlightPath(V, 'yellow')
    return P

def HidePath(P):
    global highlightID
    if highlightID:
        A.HidePath(highlightID)
        highlightID = None
    return P

FindPartition = lambda g=G, a=A: _FindPartition(g,a)

def _FindPartition(G,A):

    Q = Queue()

    for v in G.vertices:

        if dist[v] == None:
            dist[v] = 0
            Q.Append(v)
            V.AddVertex(v)

            while Q.IsNotEmpty():
                v = Q.Top()

                for w in G.Neighborhood(v):

                    if dist[w] == None:
                        dist[w] = dist[v] + 1
                        Q.Append(w)

                        if dist[w] % 2 == 1:
                            W.AddVertex(w)
                        else:
                            V.AddVertex(w)

    return (V, W)


ShowVertexCover = lambda v,w,a=A: _ShowVertexCover(v,w,a)

def _ShowVertexCover(V,W,A):

    for v in V.vertices:
        if not pred[v]:
            A.SetVertexFrameWidth(v,6)
        else:
            A.SetVertexFrameWidth(v,2)
    for v in W.vertices:
        if pred[v]:
            A.SetVertexFrameWidth(v,6)
        else:
            A.SetVertexFrameWidth(v,2)

#### Lambdas #################################################

Neighborhood    = lambda v, g=G, a=A: AnimatedNeighborhood(a,g,v,["blue"])
MatchingPartner = lambda v: M.Neighborhood(v)[0]

#### Internal initialisation #################################

for v in G.vertices:
    dist[v] = None
    pred[v] = None

A.RegisterGraphInformer(MatchingGraphInformer(G))


#### Functions Migrated from Algo file #######################

def FindExposedVertex(V,M):
    DeleteLabels()
    for v in V.Vertices():
        if not v in M.Vertices():
            pred[v] = v
            Q.Append(v)

    while Q.IsNotEmpty():
        v = Q.Top()
        for w in Neighborhood(v):
            if not pred[w]:
                pred[w] = v
                if w in M.Vertices():
                    u = MatchingPartner(w)
                    pred[u] = w
                    Q.Append(u)
                else:
                    return w
    return None

def FindAugmentingPath(V1,V2,M):
    P = []
    w = FindExposedVertex(V1,M)
    if w:
        while pred[w] != w:
            P.append((pred[w],w))
            w = pred[w]
    return ShowPath(P)

def Augment(P,M):
    for (v,w) in P:
        if (v,w) in M.Edges():
            M.DeleteEdge((v,w))
        elif (w,v) in M.Edges():
            M.DeleteEdge((w,v))
        else:
            M.AddEdge((v,w))
    return HidePath(P)