##############################################################################
#
#       This is part of CATBox (Combinatorial Algorithm Toolbox)
#       version 1.2.7 from 1/25/2021. You can find more information at
#       http://schliep.org/CATBox/
#
#       file:   WeightedMatching.pro
#       author: Winfried Hochstaettler (winfried.hochstaettler@fernuni-hagen.de)
#               Alexander Schliep (alexander@schliep.org)
#
#       Copyright (C) 1998-2020, Winfried Hochstaettler and Alexander Schliep.
#	(C) 2010 Springer Verlag.
#       
#	This is part of the Springer textbook "CATBox - an interactive course in 
#       discrete optimization".
#
#       All rights reserved. Do not distribute without written permission. 
#
#
#
#
################################################################################

#
#
#
#
#### Options #################################################
import copy

breakpoints = [3]
interactive = []
graphDisplays = 2
self.NeededProperties({'EvenOrder':1, 'Size': 0})
about = """<HTML>
<HEAD>
<TITLE>Weighted non-bipartite matching</TITLE>
</HEAD>
<BODY>

This algorithm finds a maximum weight matching in a complete graph
with Euclidean distances, given as set of points in the plane, by a
primal-dual algorithm.  One window visualizes the dual variables as
moats, the primal variables as matching and the edges of the graph
which have no slack.  A secondary window (shrinking) depicts the
graph, where the blossoms with non-zero dual variables are shrunken
into a single vertex. On this graph a cardinality matching is
performed as described in Chapter 8 of the book. It returns a maximum
matching and a Gallai-Edmonds decomposition into a <em> barrier, </em>
<em>odd components</em> and <em>even components</em>. This is used to
update the dual variables, see Chapter 9 of the book.


<H3>Visualization 
 </H3>

<H4>
Primary Window
</H4>

<H5>Vertex Colors</H5> <dl>
<dt><colordef color="#000096"></dt> <dd>
Unmatched vertices are dark blue.</dd> 

<dt><colordef color="#0000FF"></dt> <dd>
Matched vertices are blue.</dd> 

</dl>


<H5>Edge Colors</H5> <dl><dt><colordef color="#EEEEEE"></dt>
Initial color for edges. These edges occur if the corresponding
inequality of the dual program is tight, they do not belong to matching corresponding to the primal solution.</dd>

<dt><colordef color="blue"></dt> <dd>Color of edges belonging to the
current matching.</dd> <dt><colordef color="black"></dt> <dd>Color of
edges belonging to an old matching along an augmenting path.</dd>
</dl>


<H5>Further elements</H5> <dl> <dt><colordef color="red"></dt> <dd> The size of a red circle
indicates the value of the dual variables associated with the inner
(blue) vertex.</dd>

<dt><colordef color= "#FF4400""><colordef color= "#FF9900"><colordef color= "#FFFF00">
<colordef color= "#DDDD99"><colordef color= "#CCCC66"><colordef color= "#CCCC33">
<dt><colordef color= "#CCFF00"><colordef color= "#33FF00"><colordef color= "#99FF66">
<colordef color= "#66CC99""><colordef color= "#66CCCC"><colordef color= "#0099FF">
<colordef color= "#0033CC"><colordef color= "#330099"><colordef color= "#333399">
<colordef color= "#666699"><colordef color= "#9999CC"><colordef color= "#CCCCFF">
</dt><dd> Other colours are used for moats that show the value of non-trivial blossom variables with 3 or 5 or 7 etc. inner vertices.
</dd>
</dl>


<H4>
<a name="Secondary"> Secondary Window</a>
</H4>

This window depicts the graph, where the blossoms with non-zero dual
variables are shrunken into a single vertex. On this graph a
cardinality matching is performed as described in Chapter 8 of the
book. It returns a maximum matching and a Gallai-Edmonds decomposition
into a <em> barrier</em>, <em>odd components</em> and <em>even components</em>.

<p>
<p>

<H5>Vertex Colors</H5> <dl>
<dt><colordef color="red"></dt> <dd>
Vertices belonging to the barrier are red.</dd> 

<dt><colordef color="#000096"></dt> <dd>
Vertices in an odd component are dark blue.</dd> 

<dt><colordef color="#0000FF"></dt> <dd>
Vertices in an even component are blue.</dd> 

</dl>

<p>

<H5>Edge Colors</H5> 
<dl><dt><colordef color="#EEEEEE"></dt><dd>
Initial color for edges. These edges occur if the corresponding
inequality of the dual program is tight, they do not belong to matching corresponding to the primal solution.</dd>

<dt><colordef color="blue"></dt> <dd>Color of edges belonging to the current matching.</dd>
<dt><colordef color="black"></dt> <dd>Color of edges belonging to old matching along an augmenting path.</dd>
</dl>


</BODY></HTML>

"""


#from AnimatedDataStructures import AnimatedSubgraph
import copy
import math
import string
eps_prec = 0.01

#### Utility functions #########################################
def is_supervertex(v):
    return isinstance(v,tuple) or (isinstance(v,int) and v > G.Order())

def argmin(a):
    return a.index(min(a))

def argmin_min(a):
    m = min(a)
    return a.index(m), m

def UnmatchedVertices(G, M):
    result = set()
    for v in G.Vertices():
        if not M.QVertex(v) or M.Degree(v) == 0:
            result.add(v)

    return result

def dist_with_duals(G, u, v, w):
    d = dist(G, v, w) - u[v] - u[w]
    for key in u.duals[v]:
        if not key in u.duals[w]:
            d -= u[key]
    for key in u.duals[w]:
        if not key in u.duals[v]:
            d -= u[key]
    return d

def matched(M, v):
    return v in M.Vertices() and M.Degree(v) != 0

def matchingPartner(M, v):
    nbh = M.Neighborhood(v)
    return nbh[0]

def mindist(V, W):
    eps = 99999.99
    tight_edges = []

    #print "mindist", V, W
    #V = expand_supervertices(V)
    #W = expand_supervertices(W)

    #print "mindist between sets", V, W

    # NOTE: We need to compare all un-ordered pairs v \in V, w \in W,
    # v \neq w As V and W can contain the same supervertex we have to
    # compare all the un-ordered pairs such that v and

    for vv in V:
        vv_tmp = vv
        if not is_supervertex(vv):
            vv = (vv,)
        else:
            vv = u.expand_supervertices(vv)
        for ww in W:
            if vv_tmp == ww:
                continue
            if not is_supervertex(ww):
                ww = (ww,)
            else:
                # print "before expand vv=", vv, "ww=", ww, is_supervertex(ww)
                ww = u.expand_supervertices(ww)
            # print "after expand vv=", vv, "ww=", ww
            for v in vv:
                for w in ww:
                    if v is not w:
                        #print "dist", v, w, type(v), type(w), G.embedding.label
                        #print G, G.Vertices()
                        #print "dist", v, w, dist(G, v, w), u[v], u[w], "duals=", u.duals[v], u.duals[w]
                        d = dist_with_duals(G, u, v, w)
                        if d < eps - eps_prec:
                            if v < w:
                                tight_edges = [(v,w)]
                            else:
                                tight_edges = [(w,v)]
                            eps = d
                        elif (d - eps) < eps_prec:
                            if v < w:
                                tight_edges.append((v,w))
                            else:
                                tight_edges.append((w,v))

    tight_edges = list(set(tight_edges)) # remove duplicates
    return eps, tight_edges


def FindSuspectEdge(G, M):
    """

    """
    Q = Queue()
    pred = {}
    label = {}
    base = {}

    for v in UnmatchedVertices(G, M):
        pred[v] = v
        base[v] = v
        label[v] = 1
        Q.Append(v)

    #print "FindSuspectEdge: UnmatchedVertices", UnmatchedVertices(G, M)
    #print M.Edges()

    while Q.IsNotEmpty():
        v = Q.Top()
        for w in G.Neighborhood(v):
            if not w in pred:
                pred[w] = v
                base[w] = base[v]
                label[w] = 0
                if matched(M, w):
                    # print "FindSuspectEdge", v, w, M.Neighborhood(w)
                    u = matchingPartner(M, w)
                    pred[u] = w
                    base[u] = base[w]
                    label[u] = 1
                    Q.Append(u)
            elif pred[v] != w and label[w] != 0:
                #print "FindSuspectEdge", (v,w)
                return ((v,w), pred, base, label)
    return (None, pred, base, label) # Need to return for GallaiEdmondsDecomp

def FindAugmentingPath(G, M, pred, e):
    """Find an augmenting path containing e in the ShrinkableGraph G
       using the predecessor array pred
    """
    P = [e]
    v = e[0]
    while pred[v] != v:
        P.insert(0,(pred[v],v))
        v = pred[v]

    v = e[1]
    while pred[v] != v:
        P.append((v,pred[v]))
        v = pred[v]

    #print "FindAugmentingPath", P
    return P


class Bubble:
    """Graphical class representing bubbles. There are two types of bubbles: the ones
       around vertices and those around supervertices. In the latter case the minimal and
       initial bubble size is determined by the bubbles of the vertices it contains.
    """

    def __init__(self, graph, graph_display, dual_vars, key, vertices, offset_value, color="red"):
        self.graph = graph
        self.graph_display = graph_display
        self.nr_animation_steps = 25 #250
        self.ovals = [] # Supervertices have several ovals
        self.vertices = vertices
        self.value = 0.0
        self.offset_value = offset_value

        self.graph_display.CreateBubble(','.join([str(v) for v in vertices]), ','.join([str(o) for o in offset_value]), color)
        for v in vertices:
            pos = self.graph_display.VertexPosition(v)
            oval = self.graph_display.canvas.create_oval(pos.x, pos.y, pos.x, pos.y, 
                                             fill=color, outline=color, 
                                             tag="bubbles", width=0)
            self.graph_display.canvas.lower(oval,"vertices")
            # We now require that there are no edges in instances
            # self.graph_display.canvas.lower(oval,"edges")
            self.graph_display.canvas.lower(oval,"bubbles")

            self.graph_display.canvas.tag_bind(oval, "<Any-Leave>", self.graph_display.DefaultInfo)
            self.graph_display.canvas.tag_bind(oval, "<Any-Enter>", \
                                   lambda event, x=key: dual_vars.DualInfo(event,x))
            self.ovals.append(oval)

    def blow(self, new_val):
        size = self.tocanvas(self.value)
        if size < eps_prec:
            size = self.graph_display.zVertexRadius
        new_size = self.tocanvas(new_val)
        self.value = new_val

        delta_size = (new_size - size) / self.nr_animation_steps 
        j = 0
        self.graph_display.ResizeBubble(','.join([str(v) for v in self.vertices]), str(new_val))
        while j < self.nr_animation_steps:
            j = j + 1
            size = min(new_size, size + delta_size)
            for i, oval in enumerate(self.ovals):
                s = size + self.tocanvas(self.offset_value[i])
                pos = self.graph_display.VertexPosition(self.vertices[i])
                self.graph_display.canvas.coords(oval, pos.x-s, pos.y-s,
                                     pos.x+s, pos.y+s)
            self.graph_display.update()
        # Wait so we get another command in there and can get the animation time in JS
        self.graph_display.Wait()

    def deflate(self, new_val):
        size = self.tocanvas(self.value) 
        new_size = self.tocanvas(new_val)
        self.value = new_val

        delta_size = (size - new_size) / self.nr_animation_steps 
        j = 0
        self.graph_display.ResizeBubble(','.join([str(v) for v in self.vertices]), str(new_val))
        while j < self.nr_animation_steps:
            j = j + 1
            size = max(new_size, size - delta_size)          
            for i, oval in enumerate(self.ovals):
                s = size + self.tocanvas(self.offset_value[i])                
                pos = self.graph_display.VertexPosition(self.vertices[i])
                self.graph_display.canvas.coords(oval, pos.x-s, pos.y-s,
                                     pos.x+s, pos.y+s)
            self.graph_display.update()
        # Wait so we get another command in there and can get the animation time in JS
        self.graph_display.Wait()

    def __del__(self):
        self.graph_display.DeleteBubble(self.vertices)
        for oval in self.ovals:
            self.graph_display.canvas.delete(oval)
        

    def tocanvas(self, x):
        return (x * self.graph_display.zoomFactor) / 100.0
        

class DualVars:
    """Dictionary of dual variables: NOTE: we use wmc for getting the depth """

    def __init__(self, G, A):
        self.G = G
        self.A = A
        self.vars = {} # List of variables: either vertices or
        # supervertices are keys
        self.bubbles = {} # Corresponding bubbles
        self.bubble_num = 0 # Number increases as we add variables(bubbles)
        self.duals =  [ [] for i in range(G.Order() + 1)] # Which bubble dual variables are relevant
        self.vertices = {} # Dictionary of plain, expandend vertex lists
        self.colorList = ["#EE0000","#FF4400","#FF9900","#FFFF00",
                          "#DDDD99","#CCCC66","#CCCC33","#CCFF00","#33FF00",
                          "#99FF66","#66CC99","#66CCCC","#0099FF","#0033CC",
                          "#330099","#333399","#666699","#9999CC","#CCCCFF"]
        

    def expand_supervertices(self, sv):
        """returns expanded list of G-vertices that make up sv.
           sv might be a blossom or expanded supervertex (a list) or a single
           vertex (int).
        """
        if isinstance(sv, tuple):
            return list(sv)
        elif sv <= self.G.Order():
            return [sv]
        else:
            blossom = wmc.SG.blossom[sv]
            vertices = [self.expand_supervertices(v) for v in blossom]
            result = []
            for vs in vertices:
                result += vs
            return result

    def compute_offsets(self, vertices):
        offsets = [0.0] * len(vertices)
        for i, v in enumerate(vertices):
            offsets[i] += self.vars[v]
            for key in self.duals[v]:
                offsets[i] += self.vars[key]
        #print "expand_supervertex", vertices, offsets
        return offsets

    def add_variable(self, key, value):
        #print "Adding dual", key
        if is_supervertex(key):
            depth = wmc.depth[key]
            self.vertices[key] = self.expand_supervertices(key)
            offset_values = self.compute_offsets(self.vertices[key])
            # print "Adding dual", key, " with vertices ", self.vertices[key] 
            for v in self.vertices[key]:
                self.duals[v].append(key)
        else:
            depth = 0
            self.vertices[key] = [key]
            offset_values = [0.0]
        color = self.colorList[depth % len(self.colorList)]
        self.vars[key] = 0.0
        #print "Adding dual", key, depth, color
        self.bubbles[key] = Bubble(G, self.A, u, key, self.vertices[key], offset_values, color)
        
    def del_variable(self, key):
        #Assume key is expanded
        for v in self.vertices[key]:
            self.duals[v].remove(key)
        del self.bubbles[key]
        del self.vars[key]
        del self.vertices[key]
        
    def __setitem__(self, key, value):
        if not key in self.vars:
            self.add_variable(key, value)
        if value > self.vars[key]:
            self.bubbles[key].blow(value)
        else:
            self.bubbles[key].deflate(value)
        self.vars[key] = value

    def hasKey(self,key):
        return key in self.vars

    def __getitem__(self, key):
        return self.vars[key]

    def __str__(self):
        return string.join(map(lambda t:" %s = %.2f" % t, self.vars.iteritems()), "\n")

    def DualInfo(self,event, key):
        self.A.UpdateInfo("Dual var u[%s] = %0.2f" % (key, self.vars[key]))
            

class ShrinkableGraph(Graph):

    def __init__(self, G, animator = None):
        """animator: Needed"""
        Graph.__init__(self)
        self.directed = 0
        self.supervertex = {}
        self.blossom = {}
        self.g_order = G.Order()

        for v in G.Vertices():
            Graph.AddVertex(self)
        for e in G.Edges():
            Graph.AddEdge(self,e[0],e[1])

        if animator != None:
            self.labeling = copy.deepcopy(G.labeling)
            self.embedding = copy.deepcopy(G.embedding)
            self.edgeWeights = copy.deepcopy(G.edgeWeights)
            self.animator = animator
            self.graph_edit = animator
        else:
            self.graph_edit = self
            
    def addVertex(self, B=None):
        if B is None or self.animator is None:
            return Graph.AddVertex(self)
        else:
            # Compute position as center of blossom
            x = 0.0
            y = 0.0
            for v in B:
                x += self.embedding[v].x
                y += self.embedding[v].y
            v = Graph.AddVertex(self)
            return self.animator.AddVertex(x / len(B), y / len(B), v)

    def ShrinkBlossom(self, B):
        """For B the vertex set of an odd cycle in 'self', remove the
           vertices, add a new supervertex representing the cycle and
           return the supervertex
        """
        sv = self.addVertex(B)
        # print "ShrinkableGraph.ShrinkBlossom add supervertex", sv
        self.blossom[sv] = B
        # print "Shrinkblossom", B, " => ", sv
        
        # Add edges between sv and non-blossom neighbors of blossom
        for v in B:
            for w in wmc.SGV.Neighborhood(v):
                if not w in B and not self.QEdge(sv,w):
                    # print "ShriBlo adding", (sv,w), "with true edge", (v,w)
                    wmc.SG.AddEdge(sv,w)
                    self.graph_edit.AddEdge(sv,w)

            self.supervertex[v] = sv
               
        # Delete matching edges within blossom
        for v, w in wmc.SGM.Edges():
            if v in B and w in B:
                wmc.SGM.DeleteEdge(v,w)
            # Add matching edge to outside of blossom
            if (v in B and not w in B) or (not v in B and w in B):
                if v in B:
                    tmpv, tmpw = wmc.SGV.Edge(sv,w) 
                    wmc.SGM.AddEdge(tmpv, tmpw)
                    wmc.SGM.DeleteVertex(v)
                else:
                    tmpv, tmpw = wmc.SGV.Edge(v,sv) 
                    wmc.SGM.AddEdge(tmpv, tmpw)
                    wmc.SGM.DeleteVertex(w)
        for v in B:
            self.graph_edit.DeleteVertex(v)
        return sv

    def expand_supervertex(self, v, returnDepth = None):
        # Recursively expand to bottom
        terminate = 0
        depth = 0
        result = self.blossom[v][:]
        # print "expand_supervertex", v, result,
        while not terminate:
            terminate = 1            
            for i, w in enumerate(result):
                if self.is_supervertex(w):
                    result[i:i+1] = self.blossom[w]
                    #print result
                    terminate = 0
            depth += 1
        # print depth

        if returnDepth is None:
            return tuple(result)
        else:
            return (tuple(result), depth)

    def is_supervertex(self,v):
        return v > self.g_order

    def BlossomContaining(self, e, pred):
        """Return the vertex set of the blossom containing edge e give the
           predecessor array pred
        """
        v, w = e
        blossom = []
        
        p0 = [v]
        p1 = [w]

        for p in [p0, p1]:
            while pred[p[-1]] != p[-1]:
                p.append(pred[p[-1]])

        for i, v in enumerate(p0):
            if v in p1:
                break
        j = p1.index(v)

        p1slice = p1[0:j] # Need the first common just once
        p1slice.reverse()
        blossom = p0[0:i+1] + p1slice
        #print "ShrinkableGraph.BlossomContaining", blossom #p0, p1, blossom
        #print "pred=",pred
        return blossom

    def Augment(self, M, P):
        # print "Augment: Path", P
        # Unmatch first
        i = 0
        for (v,w) in P:
            if i % 2:
                for (tv, tw) in self.getTrueEdges(v, w):
                    if M.QEdge(tv, tw):
                        (tv, tw) = M.Edge(tv, tw)
                        M.DeleteEdge(tv, tw)
                        break
                    
                # Delete(v,w) in matching of SGV (SGM defined in Gallai-Edmonds: expose?)
                v,w = wmc.SGM.Edge(v,w)
                wmc.SGM.DeleteEdge(v,w)
            i += 1
        
        # match second
        i = 0
        for (v,w) in P:
            if i % 2 == 0:
                v, w = wmc.SG.Edge(v, w)
                tv, tw = self.findNextLevelEdge(v, w)
                ttv, ttw = self.findTrueEdge(v, w)

                if matched(M, ttv):
                    #print "HypoMatch",tv,(tv,tw)
                    self.HypoMatch(M, tv, (tv, tw))
                    
                if matched(M, ttw):
                    if w != tw:
                        self.HypoMatch(M, tw, (tv, tw))
                    else:
                        self.HypoMatch(M, ttw, (ttv, ttw))

                # print "Augment adding:", (tv,tw), "in G", (v,w), "in SG"
                ttv, ttw = self.findTrueEdge(v,w)
                M.AddEdge(ttv, ttw)
                wmc.SGM.AddEdge(v, w)
                #Add (v,w) to SGM

                for vv in (v,w):
                    #print "Augment expanding considering", vv, self.is_supervertex(vv),
                    if self.is_supervertex(vv) and vv in wmc.sv_added:
                        #print self.blossom[vv]
                        wmc.Unshrink(self.blossom[vv])
                        del(self.blossom[vv])
                    #print
                   
            i += 1
        # print "return from Augment M=", M.Edges(), "SGM=", wmc.SGM.Edges()


    def HypoMatch(self, M, v, entry_edge):
        """Hypomatch inside a blossom given the entry vertex v
           which should be part of a blossom"""
        if not v in self.supervertex:
            #print "HypoMatch", v, "is no supervertex"
            return
        sv = self.supervertex[v]
        b = self.blossom[sv]
        vi = b.index(v)
        #print 'HypoMatch v=', v, 'entry_edge=',entry_edge, "blossom=", b

        te1 = self.getTrueEdges(v, b[(vi + 1) % len(b)])
        te2 = self.getTrueEdges(v, b[(vi - 1) % len(b)])
        
        #print "Hypo",(tv1,tw1), M.QEdge(tv1,tw1), (tv2,tw2), M.QEdge(tv2,tw2)

        if any([M.QEdge(tv, tw) for (tv, tw) in te1]): # walk forward in blossom
            for i in range(len(b)):
                wi = (vi + 1) % len(b)
                tv, tw = self.findNextLevelEdge(b[vi], b[wi])
                #print "walk forward blos:", (b[vi], b[wi]), "next=", (tv, tw)
                
                if i % 2 == 0:
                    matching_edges = [e for e in self.getTrueEdges(b[vi], b[wi]) if M.QEdge(e[0], e[1])]
                    if matching_edges == []:
                        break
                    
                    e = matching_edges[0]
                    (ttv, ttw) = M.Edge(e[0], e[1])
                    M.DeleteEdge(ttv, ttw)
                else:
                    (firstv, firstw) = self.findNextLevelEdge(b[vi], b[wi])
                    self.HypoMatch(M, firstv, (firstv, firstw))
                    self.HypoMatch(M, firstw, (firstv, firstw))
                    
                    ttv, ttw = self.findTrueEdge(b[vi], b[wi])
                    #print "walk forward adding:", (firstv, firstw), (ttv,ttw)
                    M.AddEdge(ttv,ttw)
                vi = (vi + 1) % len(b)
            #print

        elif any([M.QEdge(tv, tw) for (tv, tw) in te2]): # walk backward in blossom
            tmp = list(range(len(b)))
            tmp.reverse()
            for i in tmp:
                wi = (vi - 1) % len(b)
                tv, tw = self.findNextLevelEdge(b[vi], b[wi])
                #print "walk backward blos:", (b[vi], b[wi]), "next=", (tv, tw)

                if i % 2 == 0:
                    matching_edges = [e for e in self.getTrueEdges(b[vi], b[wi]) if M.QEdge(e[0], e[1])]
                    if matching_edges == []:
                        break
                    
                    e = matching_edges[0]
                    (ttv, ttw) = M.Edge(e[0], e[1])
                    M.DeleteEdge(ttv, ttw)
                    
                else:

                    (firstv, firstw) = self.findNextLevelEdge(b[vi], b[wi])
                    self.HypoMatch(M, firstv, (firstv, firstw))
                    self.HypoMatch(M, firstw, (firstv, firstw))

                    ttv, ttw = self.findTrueEdge(b[vi], b[wi])
                    #print "walk backward adding:", (firstv, firstw), (ttv,ttw)
                    M.AddEdge(ttv,ttw)
                vi = (vi - 1) % len(b)

        # print "Hypo resolving super",v,"e=",entry_edge
        if self.is_supervertex(v):
            e = entry_edge
            while v in e:
                e = self.findNextLevelEdge(e[0],e[1])
                # print "Hypo resolving super",v,"e=",e,e[0] in self.blossom[v], e[1] in self.blossom[v]
            if e[0] in self.blossom[v]:
                self.HypoMatch(M,e[0],(e[0],e[1]))
            else:
                self.HypoMatch(M,e[1],(e[0],e[1]))

        # print "HypoMatch done with", v


    def findEdge(self,v,w):
        while v in self.supervertex:
            v = self.supervertex[v]
        while w in self.supervertex:
            w = self.supervertex[w]
        return (v,w)

    def getNextLevelEdge(self, v, w):
        (tv, tw) = self.getTrueEdges(v, w)[0]
        
        if tv != v and self.is_supervertex(v):
            while self.supervertex[tv] != v:
                tv = self.supervertex[tv]
                
        if tw != w and self.is_supervertex(w):
            while self.supervertex[tw] != w:
                tw = self.supervertex[tw]
        
        return (tv, tw) 

    def findNextLevelEdge(self, v, w):
        """ Find edge one level into the blossom structure. Make sure
            that both v and w are resolved into blossoms if they are both
            super-vertices.
        """
        return self.getNextLevelEdge(v, w)

    def getTrueEdges(self, V, W):
        if self.is_supervertex(V):
            result = []
            for v in self.blossom[V]:
                result += self.getTrueEdges(v, W)
            return result
        elif self.is_supervertex(W):
            result = []
            for w in self.blossom[W]:
                result += self.getTrueEdges(V, w)
            return result
        elif wmc.G.QEdge(V,W):
            return [(V,W)]
        else:
            return []
    
    def findTrueEdge(self, v, w):
        return self.getTrueEdges(v, w)[0]

def dist(G,v,w):   
    return math.sqrt((G.embedding[v].x - G.embedding[w].x)**2 +
                     (G.embedding[v].y - G.embedding[w].y)**2)


class WeightedMatchingContext:
    """Utility class to hold all relevant data structures and
       methods necessary for computation
    """

    def __init__(self, G, A, u, gato_context):
        global GD
        self.G = G
        self.GA = A
        for v in G.Vertices():
            u[v] = 0.0

        self.u = u # The dual variables
        self.gato = gato_context # Simply self in calling context
        # Make graph undirecet (all edges are gone)
        self.G.directed = 0
        self.M = AnimatedSubGraph(self.G, self.GA)
        

        # The shrinkable graph collects all the original vertices and all edges
        # and all supervertices and all their edges so that we can selectively
        # show and hide things without having to worry about renumbering as
        # we would for Adding and Deleting vertices to a graph
        H = copy.deepcopy(G) # Temporary
        self.gato.OpenSecondaryGraph(H, stripPath(self.gato.graphFileName) + " (shrinking)")
        self.SGVA = self.gato.GUI.secondaryGraphDisplay
        self.SG = ShrinkableGraph(H, self.SGVA) 
        # ... SGV is what the current instance of SG looks like
        self.SGV = SubGraph(self.SG)
        for v in self.SG.Vertices():
            self.SGV.AddVertex(v)
        self.SGVA.ShowGraph(self.SGV, stripPath(self.gato.graphFileName) + " (shrinking)")

        # YYY Moved from  GallaiEdmondsDecomposition
        self.SGM = AnimatedSubGraph(self.SGV, self.SGVA)
        self.SGM.directed = 0
        
        self.iter = 1

        self.depth = {} # How many levels of shrunken supervertices are contained


    def GallaiEdmondsDecomposition(self, G):
        """ Compute the Gallai-Edmonds decomposition on (non-bipartite) graph G

            Even: the even-labeled vertices G (inner nodes
            Odd:  the odd-labeled vertices of G: a list of either vertices or list of vertices
                  If shrinking has been necessary, a supernode in SG is recursively expanded
                  to a flat list of all vertices which are contained in the blossom of the
                  supervertex            Unlabeled: unlabeled vertices    
        """
        #print "#==== %d-th iteration =======================" % wmc.iter
        self.iter += 1

        #print "Before Matching in Shrinkable view SGV:", self.SGM.Edges()
        #print "Before Matching in original graph G:", self.M.Edges()

        SGV_order = self.SGV.Order() # Need for termination criterion

        self.sv_added = []

        # Find a maximal cardinality matching
        maximal = False
        sv = 0
        while not maximal:
            (e, pred, base, label) = FindSuspectEdge(self.SGV, self.SGM)
            #print "GallaiEdmondsDecomposition:FindSuspectEdge", (e, pred, base, label)
            if e == None:
                maximal = True
            else:
                if base[e[0]] == base[e[1]]: # Found a blossom
                    B = self.SG.BlossomContaining(e, pred)
                    # print e,"found blossom", e, B
                    sv = self.SG.ShrinkBlossom(B)
                    # print "added supernode", sv
                    self.sv_added.append(sv)
                else:
                    P = FindAugmentingPath(self.SGV, self.SGM, pred, e)
                    self.SG.Augment(self.M, P)
                    
              
            
        #print "Matching in Shrinkable view SGV:", self.SGM.Edges()
        #print "Matching in original graph G:", self.M.Edges()

        #print "GED", SGV_order, self.SGV.Order(), self.SGM.Size()
        if 2 * self.SGM.Size() < SGV_order: # Matching is not perfect
            Even = []
            Odd = []
            Unlabeled = []
            for v in self.SGV.Vertices():
                if v in label:
                    if label[v] % 2 == 0: # v even
                        Even.append(v)
                    else: # v odd
                        Odd.append(v)
                else:
                    Unlabeled.append(v)

            # COLOR A, B, C in shrunken graph (if non-empty)
            #print "Even", Even
            if Even: 
                v = copy.deepcopy(Even)
                self.SGVA.SetAllVerticesColor("red", vertices=v)
            if Odd: 
                v = copy.deepcopy(Odd)
                self.SGVA.SetAllVerticesColor(self.SGVA.g.cVertexDefault, vertices=v)
            if Unlabeled:
                v = copy.deepcopy(Unlabeled)
                self.SGVA.SetAllVerticesColor("blue", vertices=v)


            # replace supervertices by lists of vertices
            sets = [Even, Odd, Unlabeled]
            for s in sets:
                for i, v in enumerate(s):
                    if self.SG.is_supervertex(v):
                        sv, depth = self.SG.expand_supervertex(v, returnDepth = 1)
                        self.depth[sv] = depth
                        s[i] = sv

            #print (self.M, Even, Odd, Unlabeled)    
            return (self.M, Even, Odd, Unlabeled)
        else:
            return (self.M, [], [], [])



    def AddTightEdge(self, v, w):
        """Add a new tight edge"""
        # print "wmc.AddTightEdge", (v, w), self.G.QEdge(v, w) 
        if not self.G.QEdge(v, w):
            self.GA.AddEdge(v,w)

            # Figure out what the edge (v,w) corresponds to in SG
            # original graph

            # Case 1: Neither v nor w are members of a supervertex
            #     
            if not v in self.SG.supervertex and not w in self.SG.supervertex:
                if not self.SG.QEdge(v,w):
                    self.SG.AddEdge(v,w)
                if not self.SGV.QEdge(v,w):
                    self.SGVA.AddEdge(v,w)               
        # Moved the following out of if Block
        oldvv = vv = v
        while vv in self.SG.supervertex:
            done=FALSE
            while not done:
            # Find outermost supervertex
                try: 
                    vv = self.SG.supervertex[vv]
                    if not self.SG.QEdge(vv,w):
                        self.SG.AddEdge(vv,w)
                    if vv in self.SGV.Vertices() and w in self.SGV.Vertices() and \
                            not self.SGV.QEdge(vv,w):
                        self.SGVA.AddEdge(vv,w)
                    oldvv = vv
                except:
                    done=TRUE                  

        oldww = ww = w
        while ww in self.SG.supervertex:
            done = FALSE
            while not done:
                try:
                    ww = self.SG.supervertex[ww]
                    if not self.SG.QEdge(vv,ww):
                        self.SG.AddEdge(vv,ww)
                    if vv in self.SGV.Vertices() and ww in self.SGV.Vertices() and \
                            not self.SGV.QEdge(vv,ww):
                        self.SGVA.AddEdge(vv,ww)
                    oldww = ww
                except:
                    done = TRUE
        self.M.RaiseEdges()

    def Unshrink(self, sv):
        #print "Unshrink", sv, M.Edges()
        svset = sv
        sv = self.SG.supervertex[sv[0]] # NOTE: this uniquely maps from
        # (v_1, ..., v_k) to v with v > G.Order()
        sv = findSGsupervertex(sv)
        
        for v in self.SGV.Neighborhood(sv):
            if self.SGM.QEdge(sv,v):
                tmpv, tmpw = self.SGM.Edge(sv,v)
                self.SGM.DeleteEdge(tmpv,tmpw)

            tmpv, tmpw = self.SG.Edge(sv,v)
            self.SGVA.DeleteEdge(tmpv, tmpw)
        self.SGVA.DeleteVertex(sv)
        for v in self.SG.blossom[sv]:
            self.SGVA.AddVertex(self.SG.embedding[v].x, self.SG.embedding[v].y, v)

        # print "self.SGV.Vertices = ", self.SGV.Vertices()
        # Find all tight edges and add them
        for v in self.SGV.Vertices():
            for w in self.SGV.Vertices():
                if v < w: 

                    # print (v,w), self.SG.QEdge(v,w)
                    if self.SG.is_supervertex(v):
                        #print "Unshrink v", v, self.SG.is_supervertex(v), self.SG.blossom[v]
                        tmpv = self.SG.blossom[v]
                    else:
                        tmpv = [v]
                    if self.SG.is_supervertex(w):
                        #print "Unshrink w", w, self.SG.is_supervertex(w), self.SG.blossom[w]
                        tmpw = self.SG.blossom[w]
                    else:
                        tmpw = [w]
                    d, t = mindist(tmpv,tmpw)

                    #print "d(%d,%d) = %d" % (v,w,d), dist(G, v, w), u[v], u[w]
                    #print "d(%d,%d) = %d" % (v,w,d)
                    if d < eps_prec:
                        # print "eps_prec", v,w
                        if not self.SG.QEdge(v,w):
                            self.SG.AddEdge(v,w)
                        if not self.SGV.QEdge(v,w):
                            self.SGVA.AddEdge(v,w)

                        if not self.SGM.QEdge(v,w):
                            true_edges = self.SG.getTrueEdges(v, w)
                            if any([self.M.QEdge(tv, tw) for (tv, tw) in true_edges]):
                                self.SGM.AddEdge(v, w)

        for v in self.SG.blossom[sv]:
            del[self.SG.supervertex[v]]
        # print "nach unshrink", M.Edges(), "geshrunkene", self.SGM.Edges()

class WeightedMatchingInformer(GraphInformer):

    def __init__(self, G, wmc,u):
        GraphInformer.__init__(self,G)
        self.wmc = wmc
        self.u = u

    def EdgeInfo(self, v, w):
        l = self.G.edgeWeights[0][(v,w)]
        return  "Edge (%d,%d) has length %0.2f" % (v,w,l)

    def VertexInfo(self, v):
        return "Vertex %d, u[%d] = %0.2f" % (v,v,self.u[v])

    def DefaultInfo(self):
        return "Cardinality of matching is %d" % wmc.M.Size()


class ShrunkenInformer(GraphInformer):

    def __init__(self, G, wmc,u):
        GraphInformer.__init__(self,G)
        self.wmc = wmc
        self.u = u

    def VertexInfo(self, v):
        if self.wmc.SG.is_supervertex(v):
            return "Supervertex %d = %s" % (v,self.wmc.SG.expand_supervertex(v))
        else:
            return GraphInformer.VertexInfo(self, v)

    def EdgeInfo(self, v, w):
        tv,tw = self.wmc.SG.findTrueEdge(v,w)
        return "Edge (%d,%d) originally (%d,%d)" % (v,w,tv,tw)
        
    def DefaultInfo(self):
        deficiency = wmc.SGV.Order() - 2*wmc.SGM.Size()
        return "Number of unmatched vertices is %d" % deficiency


            
# =================== Initialisation =============================
u = DualVars(G, A)
wmc = WeightedMatchingContext(G, A, u, self)
Infinity = 999999.99
# =================== Syntactic sugar =============================

GallaiEdmondsDecomposition = lambda graph: wmc.GallaiEdmondsDecomposition(graph)
is_dualvar = lambda v: u.hasKey(v)
AddTightEdge = lambda v,w,thewmc = wmc: thewmc.AddTightEdge(v,w)
def Unshrink(sv, thewmc=wmc, theu=u):
    thewmc.Unshrink(sv)
    theu.del_variable(sv)

A.RegisterGraphInformer(WeightedMatchingInformer(G,wmc,u))
wmc.SGVA.RegisterGraphInformer(ShrunkenInformer(G,wmc,u))

def RemoveIncidentUntightEdges(v, A, B, C, thewmc=wmc):
    nbh = thewmc.SGV.Neighborhood(v)
    for i, w in enumerate(nbh):
        if is_supervertex(w):
            nbh.remove(w)
            nbh += thewmc.SG.blossom[w]
    for w in thewmc.SGV.Neighborhood(v):
        if w in C or w in A:
            d = dist_with_duals(G,u,v,w)
            if d > eps_prec:
                vv, ww = thewmc.SGV.Edge(v,w)
                thewmc.SGVA.DeleteEdge(vv,ww)
                thewmc.GA.DeleteEdge(vv,ww)


def findSGsupervertex(v, SG=wmc.SG):
    while v in SG.supervertex:
        v = SG.supervertex[v]
    return v

def RemoveIncidentUntightEdgesFromSupervertex(sv, A, B, C, thewmc=wmc):
    # sv is a tuple of G-vertices
    G = thewmc.G
    for v in sv:
        Nv = G.Neighborhood(v)
        for w in Nv:
            if w not in sv:
                d = dist_with_duals(G, thewmc.u, v, w)
                if d > eps_prec:
                    vv, ww = thewmc.G.Edge(v, w)
                    thewmc.GA.DeleteEdge(vv, ww)
                    
                    vv, ww = tuple(map(findSGsupervertex, (v, w)))
                    if thewmc.SGV.QEdge(vv, ww):
                        vv, ww = thewmc.SGV.Edge(vv, ww)
                        thewmc.SGVA.DeleteEdge(vv, ww)
                    
def CheckSolution(M, u):    
    # Check if M is a matching (primal feasibility)
    valid = True
    for e in M.Edges():
        for v in e:
            if M.Degree(v) != 1:
                print("deg(%d) == %d != 1" % (v, M.Degree(v)))
                valid = False
    if not valid:
        print("*** Invalid matching! ***")            
            
    # Check strong duality
    pweight = sum([dist(M, v, w) for (v, w) in M.Edges()])
    dweight = sum(u.vars.values())

    print("Difference between primal and dual objective function value: %g" % abs(pweight - dweight))
    if abs(pweight - dweight) < eps_prec:
        print("OK.")
    else:
        print("*** Difference too big, duality does not hold. ***")

def prettyprint_blossom(blossom, v, indent=0, delta=2):
    print(' ' * indent * delta,)
    print(v,)
    if v in blossom:
        print(': {')
        for w in blossom[v]:
            prettyprint_blossom(blossom, w, indent+1, delta)
        print(' ' * indent * delta, '}',)
    print()
